# -*- coding: utf-8 -*-
"""
gbr2laser V0.13alpha
Calculation of gcode for pcb-exposure with a UV-laser on a CNC-machine.
Datasource: gbr-export from KiCAD 4.0.2 - 4.0.3 stable
Datatarget: G-Code for LinuxCNC
Tested in python 2.7.6
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <FlorenzW> wrote this file. As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return. Florian W.
 * ----------------------------------------------------------------------------
"""

import math
from shapely.geometry import *
from shapely.affinity import *
from shapely.ops import cascaded_union
import matplotlib.pyplot as plt
from copy import deepcopy
import fpformat
from random import randint
import ConfigParser

config = ConfigParser.RawConfigParser()

ax = []              #for pyplot

xMin = 0
yMin = 0
xMax = 10
yMax = 10

action_number = 1
bidirectional = True
filename_input = 'Arduino-B.Cu.gbr'
filename_output = 'output.ngc'
filename_drl_input = 'Arduino.drl'
filename_drl_output = 'output_drill.ngc'
laser_diameter = 0.2
set_xMin_yMin_to_zero = True
border = 5.0
distance_fillinglines = 0.15
speed_perimeters = 1000
speed_filling = 1600
switch_laser_on = 'M62 P-0'
switch_laser_off = 'M63 P-0'
optimize_fillingpattern = True

export_perimeter = True
export_filling = True


def plot_drills(drills, color):
    i = 0
    while i < len(drills):
        line = []
        line.append([drills[i][1] - 0.5, drills[i][2]])
        line.append([drills[i][1] + 0.5, drills[i][2]])
        plot_line(line, color)
        line = []
        line.append([drills[i][1], drills[i][2] + 0.5])
        line.append([drills[i][1], drills[i][2] - 0.5])
        plot_line(line, color)
        i = i + 1


def export_drills(drills):
    print("")
    print("Plotting drills")
    DateiOutput = open(filename_drl_output,"w")

    DateiOutput.write("G21")
    DateiOutput.write("\n")
    DateiOutput.write("G90")
    DateiOutput.write("\n")
    DateiOutput.write("G94")
    DateiOutput.write("\n")
    DateiOutput.write("G0 Z1.0")
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0.0 Y0.0")
    DateiOutput.write("\n")
    DateiOutput.write("F50")
    DateiOutput.write("\n")

    i = 0
    while i < len(drills):
        DateiOutput.write("G0 X")
        DateiOutput.write(fpformat.fix(drills[i][1], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(drills[i][2], 3))
        DateiOutput.write(' ; ')
        DateiOutput.write(drills[i][0])
        DateiOutput.write("\n")
        DateiOutput.write("G1 Z-2.50")
        DateiOutput.write("\n")
        DateiOutput.write("G1 Z1")
        DateiOutput.write("\n")
        
        i = i + 1

    DateiOutput.write("G0 X0 Y0")
    DateiOutput.write("\n")
    DateiOutput.write("M2")
    DateiOutput.write("\n")
    DateiOutput.close()

def export_x_inv_drills(drills):
    print("")
    print("Plotting drills x_inv")
    filename_drl_inv = filename_drl_output.split(".")[0] + '_invX.' + filename_drl_output.split(".")[1]
    DateiOutput = open(filename_drl_inv,"w")

    DateiOutput.write("G21")
    DateiOutput.write("\n")
    DateiOutput.write("G90")
    DateiOutput.write("\n")
    DateiOutput.write("G94")
    DateiOutput.write("\n")
    DateiOutput.write("G0 Z1.0")
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0.0 Y0.0")
    DateiOutput.write("\n")
    DateiOutput.write("F50")
    DateiOutput.write("\n")

    i = 0
    while i < len(drills):
        DateiOutput.write("G0 X")
        DateiOutput.write(fpformat.fix(drills[i][1] * (-1), 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(drills[i][2], 3))
        DateiOutput.write(' ; ')
        DateiOutput.write(drills[i][0])
        DateiOutput.write("\n")
        DateiOutput.write("G1 Z-2.50")
        DateiOutput.write("\n")
        DateiOutput.write("G1 Z1")
        DateiOutput.write("\n")
        
        i = i + 1

    DateiOutput.write("G0 X0 Y0")
    DateiOutput.write("\n")
    DateiOutput.write("M2")
    DateiOutput.write("\n")
    DateiOutput.close()


def export_y_inv_drills(drills):
    print("")
    print("Plotting drills y_inv")
    filename_drl_inv = filename_drl_output.split(".")[0] + '_invY.' + filename_drl_output.split(".")[1]
    DateiOutput = open(filename_drl_inv,"w")

    DateiOutput.write("G21")
    DateiOutput.write("\n")
    DateiOutput.write("G90")
    DateiOutput.write("\n")
    DateiOutput.write("G94")
    DateiOutput.write("\n")
    DateiOutput.write("G0 Z1.0")
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0.0 Y0.0")
    DateiOutput.write("\n")
    DateiOutput.write("F50")
    DateiOutput.write("\n")

    i = 0
    while i < len(drills):
        DateiOutput.write("G0 X")
        DateiOutput.write(fpformat.fix(drills[i][1], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(drills[i][2] * (-1), 3))
        DateiOutput.write(' ; ')
        DateiOutput.write(drills[i][0])
        DateiOutput.write("\n")
        DateiOutput.write("G1 Z-2.50")
        DateiOutput.write("\n")
        DateiOutput.write("G1 Z1")
        DateiOutput.write("\n")
        
        i = i + 1

    DateiOutput.write("G0 X0 Y0")
    DateiOutput.write("\n")
    DateiOutput.write("M2")
    DateiOutput.write("\n")
    DateiOutput.close()


def translate_drills(drills, x, y):
    i = 0
#    print(drills)
    while i < len(drills):
#        print(x)
#        print(y)
#        print(drills[i])
        drills[i][1] = drills[i][1] + x;
        drills[i][2] = drills[i][2] + y;
#        print(drills[i])
        i = i + 1
    return(drills)


def import_drill(fname):
    file_length = file_len(fname)
    line = " "
    drills = []

    print("")
    print("Import drills"),
    print(fname)

    file_input = open(fname)

    i = 0
    while ((i < file_length) and line[0] <> 'T'):
        line = file_input.readline()
        line = line.strip()   
        i = i + 1
    
    while ((i < file_length) and line[0] <> '%'):
        drill = line.split('C')
        drills.append(drill)
        line = file_input.readline()
        line = line.strip()
        i = i + 1

#    print(drills)

    while ((i < file_length) and line[0] <> 'T'):
        line = file_input.readline()
        line = line.strip()   
        i = i + 1

    holes = []
    tool = line
    line = file_input.readline()
    line = line.strip()   
    i = i + 1

    while ((i < file_length) and line[0:1] <> 'T0'):
        if line[0] == "T":
            tool = line
        if line[0] == "X":
            hole = []
            hole.append(tool)
            hole.append(float(line.split('X')[1].split('Y')[0]))
            hole.append(float(line.split('X')[1].split('Y')[1]))
            holes.append(hole)
        
        line = file_input.readline()
        line = line.strip()   
        i = i + 1
#    print(holes)
    return(holes)


def read_config():
    global action_number
    global bidirectional
    global filename_input
    global filename_output
    global filename_drl_input
    global filename_drl_output
    global laser_diameter
    global set_xMin_yMin_to_zero
    global border
    global distance_fillinglines
    global speed_perimeters
    global speed_filling
    global switch_laser_on
    global switch_laser_off
    global optimize_fillingpattern

    config = ConfigParser.RawConfigParser()
    config.read('config.cfg')

    action_number = config.getint('gbr2laser', 'action_number')
    bidirectional = config.getboolean('gbr2laser', 'bidirectional')
    filename_input = config.get('gbr2laser', 'filename_input')
    filename_output = config.get('gbr2laser', 'filename_output')
    filename_drl_input = config.get('gbr2laser', 'filename_drl_input')
    filename_drl_output = config.get('gbr2laser', 'filename_drl_output')
    laser_diameter = config.getfloat('gbr2laser', 'laser_diameter')
    set_xMin_yMin_to_zero = config.getboolean('gbr2laser', 'set_xMin_yMin_to_zero')
    border = config.getfloat('gbr2laser', 'border')
    distance_fillinglines = config.getfloat('gbr2laser', 'distance_fillinglines')
    speed_perimeters = config.getint('gbr2laser', 'speed_perimeters')
    speed_filling = config.getint('gbr2laser', 'speed_filling')
    switch_laser_on = config.get('gbr2laser', 'switch_laser_on')
    switch_laser_off = config.get('gbr2laser', 'switch_laser_off')
    optimize_fillingpattern = config.getboolean('gbr2laser', 'optimize_fillingpattern')


def write_config():
    global action_number
    global bidirectional
    global filename_input
    global filename_output
    global filename_drl_input
    global filename_drl_output
    global laser_diameter
    global set_xMin_yMin_to_zero
    global border
    global distance_fillinglines
    global speed_perimeters
    global speed_filling
    global switch_laser_on
    global switch_laser_off
    global optimize_fillingpattern
    
    config.add_section('gbr2laser')
    config.set('gbr2laser', 'action_number', action_number)
    config.set('gbr2laser', 'bidirectional', bidirectional)
    config.set('gbr2laser', 'filename_input', filename_input)
    config.set('gbr2laser', 'filename_output', filename_output)
    config.set('gbr2laser', 'filename_drl_input', filename_drl_input)
    config.set('gbr2laser', 'filename_drl_output', filename_drl_output)
    config.set('gbr2laser', 'laser_diameter', laser_diameter)
    config.set('gbr2laser', 'set_xMin_yMin_to_zero', set_xMin_yMin_to_zero)
    config.set('gbr2laser', 'border', border)
    config.set('gbr2laser', 'distance_fillinglines', distance_fillinglines)
    config.set('gbr2laser', 'speed_perimeters', speed_perimeters)
    config.set('gbr2laser', 'speed_filling', speed_filling)
    config.set('gbr2laser', 'switch_laser_on', switch_laser_on)
    config.set('gbr2laser', 'switch_laser_off', switch_laser_off)
    config.set('gbr2laser', 'optimize_fillingpattern', optimize_fillingpattern)

    with open('config.cfg', 'wb') as configfile:
        config.write(configfile)


def ask_config():
    global action_number
    global bidirectional
    global filename_input
    global filename_output
    global filename_drl_input
    global filename_drl_output
    global laser_diameter
    global set_xMin_yMin_to_zero
    global border
    global distance_fillinglines
    global speed_perimeters
    global speed_filling
    global switch_laser_on
    global switch_laser_off
    global optimize_fillingpattern

    print("What do you want to calculate:")
    print("1: Perimeter + filling")
    print("2: Only filling")
    print("3: Only perimeter")
    action_number = int(raw_input('(default ' + str(action_number) + ') ') or (action_number))

    filename_input = raw_input('Gerber-File as input(default ' + filename_input + ') ') or (filename_input)

    filename_output = raw_input('ngc-File as output (default ' + filename_output + ')') or (filename_output)

    filename_drl_input = raw_input('Drillfile as input(default ' + filename_drl_input + ') ') or (filename_drl_input)

    filename_drl_output = raw_input('ngc-File as drilloutput (default ' + filename_drl_output + ')') or (filename_drl_output)

    laser_diameter = float(raw_input('Assumed laserdiameter in mm (default ' + str(laser_diameter) + ')') or (laser_diameter))

    if set_xMin_yMin_to_zero:
        if (raw_input("Set xMin and yMin to zero Y/N (default Y)") or ("Y")) == "Y":
            set_xMin_yMin_to_zero = True
        else:
            set_xMin_yMin_to_zero = False
    else:
        if (raw_input("Set xMin and yMin to zero Y/N (default N)") or ("N")) == "Y":
            set_xMin_yMin_to_zero = True
        else:
            set_xMin_yMin_to_zero = False
            
    border = float(raw_input('Border around your layout in mm (default ' + str(border) + ')') or (border))
        
    distance_fillinglines = float(raw_input('Distance between filling lines im mm (default ' + str(distance_fillinglines) + ')') or (distance_fillinglines))

    speed_perimeters = int(raw_input('Perimeterspeed in mm/min (default ' + str(speed_perimeters) + ')') or (speed_perimeters))
    speed_filling = int(raw_input('Fillingspeed in mm/min (default ' + str(speed_filling) + ')') or (speed_filling))

    #optimizing_cycles = int(raw_input("Number of optimizing cycles to reduce positioning (default 20)") or ("20"))

    switch_laser_on = raw_input('G-Code to switch laser on (default ' + switch_laser_on + ')') or (switch_laser_on)
    switch_laser_off = raw_input('G-Code to switch laser on (default ' + switch_laser_off + ')') or (switch_laser_off)

    if optimize_fillingpattern:
        if (raw_input("Optimize fillingpattern Y/N (default Y)") or ("Y")) == "Y":
            optimize_fillingpattern = True
        else:
            optimize_fillingpattern = False
    else:
        if (raw_input("Optimize fillingpattern Y/N (default N)") or ("N")) == "Y":
            optimize_fillingpattern = True
        else:
            optimize_fillingpattern = False

    if optimize_fillingpattern == True:
        bidirectional = True

    if (action_number <> 3) and (optimize_fillingpattern == False):
        if bidirectional:
            if (raw_input('Filling bidirectional Y/N (default Y) ') or ('Y')) == 'Y':
                bidirectional = True
            else:
                bidirectional = False
        else:
            if (raw_input('Filling bidirectional Y/N (default N) ') or ('N')) == 'Y':
                bidirectional = True
            else:
                bidirectional = False


def calculate_vector(line):
    vector = []
    vector.append(line[1][0]-line[0][0])
    vector.append(line[1][1]-line[0][1])
    length = line_length([[0,0], vector])

    vector[0] = vector[0] / length
    vector[1] = vector[1] / length

    return vector


def shrink_lines(filling_pattern, laser_diameter):
    i = 0
    while i < len(filling_pattern):
        vector = calculate_vector(filling_pattern[i])
        filling_pattern[i][0][0] = filling_pattern[i][0][0] + vector[0] * laser_diameter * 0.25
        filling_pattern[i][0][1] = filling_pattern[i][0][1] + vector[1] * laser_diameter * 0.25

        filling_pattern[i][1][0] = filling_pattern[i][1][0] - vector[0] * laser_diameter * 0.25
        filling_pattern[i][1][1] = filling_pattern[i][1][1] - vector[1] * laser_diameter * 0.25
        i = i + 1

def del_short_lines(filling_pattern, laser_diameter):
    i = 0
    while i < len(filling_pattern):
        if line_length(filling_pattern[i]) < laser_diameter:
            del filling_pattern[i]
            i = i - 1
        i = i + 1

def optimizing_outline_travel_V2(perimeter):
    all_polylines = []
    i = 0
    while i < len(perimeter):
        j = 0
        polyline = []
        while j < len(perimeter[i].coords):
            polyline.append(perimeter[i].coords[j])
            j = j + 1
        all_polylines.append(polyline)
        i = i + 1

    print("")    
    print("Optimizing outline arrangement")
    print("Positioninglength old: "),
    print(calculate_positioning_length_outlines(all_polylines))

    all_polylines_new = []
    punkt_old = [0, 0]

    while len(all_polylines) > 0:
        i = 0
        length_old = line_length([punkt_old, all_polylines[i][0]])
        linenumber = 0
        while i < len(all_polylines):
            length_new = line_length([punkt_old, all_polylines[i][0]])
            if length_new < length_old:
                length_old = length_new
                linenumber = i
            i = i + 1

        punkt_old = deepcopy(all_polylines[linenumber][0])
        all_polylines_new.append(all_polylines[linenumber])
        del all_polylines[linenumber]

    print("Positioninglength new: "),
    print(calculate_positioning_length_outlines(all_polylines_new))

    perimeter = []
    i = 0
    while i < len(all_polylines_new):
        perimeter.append(all_polylines_new[i])
        i = i + 1

    return(MultiLineString(perimeter))
            


def optimizing_outline_arrangement(perimeter):
    global optimizing_cycles
    all_polylines = []
    i = 0
    while i < len(perimeter):
        j = 0
        polyline = []
        while j < len(perimeter[i].coords):
            polyline.append(perimeter[i].coords[j])
            j = j + 1
        all_polylines.append(polyline)
        i = i + 1

    print("")    
    print("Optimizing outline arrangement, this takes a while...")
    print("Positioninglength old: "),
    print(calculate_positioning_length_outlines(all_polylines))

    t = 0
    while t < optimizing_cycles:

        i = 0
        while i < 1000:

            length_old = calculate_positioning_length_outlines(all_polylines)
            random_posA = randint(0,len(all_polylines)-1)
            random_posB = randint(0,len(all_polylines)-1)
            all_polylines[random_posB], all_polylines[random_posA] = all_polylines[random_posA], all_polylines[random_posB]
            if length_old <= calculate_positioning_length_outlines(all_polylines): 
                all_polylines[random_posB], all_polylines[random_posA] = all_polylines[random_posA], all_polylines[random_posB]
            i = i + 1

        t = t + 1
        print("Cycle: "),
        print(t),
        print("  "),
        print(length_old)


    print("")      
    print("Positioninglength new: "),
    print(calculate_positioning_length_outlines(all_polylines))

    perimeter = []
    i = 0
    while i < len(all_polylines):
        perimeter.append(all_polylines[i])
        i = i + 1

    return(MultiLineString(perimeter))


def calculate_positioning_length_outlines(all_polylines):
    i = 0
    length = 0
    while i < len(all_polylines)-1:
        length = length + line_length([all_polylines[i][0],all_polylines[i+1][0]])
        i = i + 1
    return length


#Returns the length of a line
def line_length(line):
    return math.sqrt(pow(line[0][0]-line[1][0],2) + pow(line[0][1]-line[1][1],2))


def export_gcode(geometriy_outlines, filling_pattern, export_perimeter, export_filling):
    print("")
    print("Exporting GCode")
    DateiOutput = open(filename_output,"w")

    DateiOutput.write("G21")
    DateiOutput.write("\n")
    DateiOutput.write("G90")
    DateiOutput.write("\n")
    DateiOutput.write("G94")
    DateiOutput.write("\n")
    DateiOutput.write("G64 P0.001")
    DateiOutput.write("\n")
#    DateiOutput.write("M5")
#    DateiOutput.write("M63 P-0")
    DateiOutput.write(switch_laser_off)
    DateiOutput.write("\n")
    DateiOutput.write("F")
    DateiOutput.write(str(speed_perimeters))
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0.0 Y0.0 Z0.0")
    DateiOutput.write("\n")

    if export_perimeter:
        i = 0
        while (i < len(geometriy_outlines)):

            line = LineString(geometriy_outlines[i])
            x, y = line.coords.xy

            DateiOutput.write("G0 X")
            DateiOutput.write(fpformat.fix(x[0], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(y[0], 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M3")
#            DateiOutput.write("M62 P-0")
            DateiOutput.write(switch_laser_on)
            DateiOutput.write("\n")

            j = 1
            while (j < len(geometriy_outlines[i].coords)-1):
                DateiOutput.write("G1 X")
                DateiOutput.write(fpformat.fix(x[j], 3))
                DateiOutput.write(" Y")
                DateiOutput.write(fpformat.fix(y[j], 3))
                DateiOutput.write("\n")
                j = j + 1

            DateiOutput.write("G1 X")
            DateiOutput.write(fpformat.fix(x[j], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(y[j], 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M5")
#            DateiOutput.write("M63 P-0")
            DateiOutput.write(switch_laser_off)
            DateiOutput.write("\n")

            i = i + 1    
            
    DateiOutput.write("F"),
    DateiOutput.write(str(speed_filling))
    DateiOutput.write("\n")

    if export_filling:
        i = 0
        while i < len(filling_pattern):
            DateiOutput.write("G0 X")
            DateiOutput.write(fpformat.fix(filling_pattern[i][0][0], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(filling_pattern[i][0][1], 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M3")
#            DateiOutput.write("M62 P-0")
            DateiOutput.write(switch_laser_on)
            DateiOutput.write("\n")
            DateiOutput.write("G1 X")
            DateiOutput.write(fpformat.fix(filling_pattern[i][1][0], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(filling_pattern[i][1][1], 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M5")
#            DateiOutput.write("M63 P-0")
            DateiOutput.write(switch_laser_off)
            DateiOutput.write("\n")

            i = i + 1

    DateiOutput.write("G0 X0 Y0")
    DateiOutput.write("\n")
    DateiOutput.write("M2")
    DateiOutput.write("\n")
    DateiOutput.close()


def export_x_inv_gcode(geometriy_outlines, filling_pattern, export_perimeter, export_filling):
    print("")
    print("Exporting GCode x_inv")
    filename_inv = filename_output.split(".")[0] + '_invX.' + filename_output.split(".")[1]
    DateiOutput = open(filename_inv,"w")

    DateiOutput.write("G21")
    DateiOutput.write("\n")
    DateiOutput.write("G90")
    DateiOutput.write("\n")
    DateiOutput.write("G94")
    DateiOutput.write("\n")
    DateiOutput.write("G64 P0.001")
    DateiOutput.write("\n")
#    DateiOutput.write("M5")
#    DateiOutput.write("M63 P-0")
    DateiOutput.write(switch_laser_off)
    DateiOutput.write("\n")
    DateiOutput.write("F")
    DateiOutput.write(str(speed_perimeters))
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0.0 Y0.0 Z0.0")
    DateiOutput.write("\n")

    if export_perimeter:
        i = 0
        while (i < len(geometriy_outlines)):

            line = LineString(geometriy_outlines[i])
            x, y = line.coords.xy

            DateiOutput.write("G0 X")
            DateiOutput.write(fpformat.fix(x[0]*(-1), 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(y[0], 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M3")
#            DateiOutput.write("M62 P-0")
            DateiOutput.write(switch_laser_on)
            DateiOutput.write("\n")

            j = 1
            while (j < len(geometriy_outlines[i].coords)-1):
                DateiOutput.write("G1 X")
                DateiOutput.write(fpformat.fix(x[j]*(-1), 3))
                DateiOutput.write(" Y")
                DateiOutput.write(fpformat.fix(y[j], 3))
                DateiOutput.write("\n")
                j = j + 1

            DateiOutput.write("G1 X")
            DateiOutput.write(fpformat.fix(x[j]*(-1), 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(y[j], 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M5")
#            DateiOutput.write("M63 P-0")
            DateiOutput.write(switch_laser_off)
            DateiOutput.write("\n")

            i = i + 1    
            
    DateiOutput.write("F"),
    DateiOutput.write(str(speed_filling))
    DateiOutput.write("\n")

    if export_filling:
        i = 0
        while i < len(filling_pattern):
            DateiOutput.write("G0 X")
            DateiOutput.write(fpformat.fix(filling_pattern[i][0][0]*(-1), 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(filling_pattern[i][0][1], 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M3")
#            DateiOutput.write("M62 P-0")
            DateiOutput.write(switch_laser_on)
            DateiOutput.write("\n")
            DateiOutput.write("G1 X")
            DateiOutput.write(fpformat.fix(filling_pattern[i][1][0]*(-1), 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(filling_pattern[i][1][1], 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M5")
#            DateiOutput.write("M63 P-0")
            DateiOutput.write(switch_laser_off)
            DateiOutput.write("\n")

            i = i + 1

    DateiOutput.write("G0 Z10")
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0 Y0")
    DateiOutput.write("\n")
    DateiOutput.write("M2")
    DateiOutput.write("\n")
    DateiOutput.close()


def export_y_inv_gcode(geometriy_outlines, filling_pattern, export_perimeter, export_filling):
    print("")
    print("Exporting GCode y_inv")
    filename_inv = filename_output.split(".")[0] + '_invY.' + filename_output.split(".")[1]
    DateiOutput = open(filename_inv,"w")

    DateiOutput.write("G21")
    DateiOutput.write("\n")
    DateiOutput.write("G90")
    DateiOutput.write("\n")
    DateiOutput.write("G94")
    DateiOutput.write("\n")
    DateiOutput.write("G64 P0.001")
    DateiOutput.write("\n")
#    DateiOutput.write("M5")
#    DateiOutput.write("M63 P-0")
    DateiOutput.write(switch_laser_off)
    DateiOutput.write("\n")
    DateiOutput.write("F")
    DateiOutput.write(str(speed_perimeters))
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0.0 Y0.0 Z0.0")
    DateiOutput.write("\n")

    if export_perimeter:
        i = 0
        while (i < len(geometriy_outlines)):

            line = LineString(geometriy_outlines[i])
            x, y = line.coords.xy

            DateiOutput.write("G0 X")
            DateiOutput.write(fpformat.fix(x[0], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(y[0]*(-1), 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M3")
#            DateiOutput.write("M62 P-0")
            DateiOutput.write(switch_laser_on)
            DateiOutput.write("\n")

            j = 1
            while (j < len(geometriy_outlines[i].coords)-1):
                DateiOutput.write("G1 X")
                DateiOutput.write(fpformat.fix(x[j], 3))
                DateiOutput.write(" Y")
                DateiOutput.write(fpformat.fix(y[j]*(-1), 3))
                DateiOutput.write("\n")
                j = j + 1

            DateiOutput.write("G1 X")
            DateiOutput.write(fpformat.fix(x[j], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(y[j]*(-1), 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M5")
#            DateiOutput.write("M63 P-0")
            DateiOutput.write(switch_laser_off)
            DateiOutput.write("\n")

            i = i + 1    
            
    DateiOutput.write("F"),
    DateiOutput.write(str(speed_filling))
    DateiOutput.write("\n")

    if export_filling:
        i = 0
        while i < len(filling_pattern):
            DateiOutput.write("G0 X")
            DateiOutput.write(fpformat.fix(filling_pattern[i][0][0], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(filling_pattern[i][0][1]*(-1), 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M3")
#            DateiOutput.write("M62 P-0")
            DateiOutput.write(switch_laser_on)
            DateiOutput.write("\n")
            DateiOutput.write("G1 X")
            DateiOutput.write(fpformat.fix(filling_pattern[i][1][0], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(filling_pattern[i][1][1]*(-1), 3))
            DateiOutput.write("\n")
#            DateiOutput.write("M5")
#            DateiOutput.write("M63 P-0")
            DateiOutput.write(switch_laser_off)
            DateiOutput.write("\n")

            i = i + 1

    DateiOutput.write("G0 Z10")
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0 Y0")
    DateiOutput.write("\n")
    DateiOutput.write("M2")
    DateiOutput.write("\n")
    DateiOutput.close()


def plot_fillingpattern(filling_pattern, color):
    i = 0
    while i < len(filling_pattern):
        line = []
        line.append(filling_pattern[i][0])
        line.append(filling_pattern[i][1])
        plot_line(line, color)
        i = i + 1


def plot_fillingpattern_positioning(filling_pattern, color):
    i = 1
    while i < len(filling_pattern):
        line = []
        line.append(filling_pattern[i-1][1])
        line.append(filling_pattern[i][0])
        plot_line(line, color)
        i = i + 1


def optimize_fillingpattern_positioning(filling_pattern):
    i = 0
    while i < len(filling_pattern) - 2:
        j = i + 1
        positioning_length = line_length([filling_pattern[i][1],filling_pattern[j][0]])
        positioning_pos = j

        while j < len(filling_pattern) - 1:
            j = j + 1
            if line_length([filling_pattern[i][1],filling_pattern[j][0]]) < positioning_length:
                positioning_length = line_length([filling_pattern[i][1],filling_pattern[j][0]])
                positioning_pos = j
        
#        print(positioning_length, '  ', positioning_pos)

        element = filling_pattern[i+1]
        filling_pattern[i+1] = filling_pattern[positioning_pos]
        filling_pattern[positioning_pos] = element

        i = i + 1
        

def calculate_filling_pattern(all_lines, bidirectional):
    print("")
    print("Calculating fillingpattern - this takes a while...")
    global xMin
    global xMax
    global yMin
    global yMax
    global distance_fillinglines
    x = xMin
    fillingpattern = []
    filling_orientation = True   #is switched every line to turn the orientation

    while x < xMax:
        i = 0
        intersections = []
        while i < len(all_lines):

            pointK = []
            pointK.append(x)
            pointK.append(0)
            pointL = []
            pointL.append(x)
            pointL.append(1)
            pointM = []
            pointM = all_lines[i][0]
            pointN = []
            pointN = all_lines[i][1]
            
            if ((pointM[0] < x) and (pointN[0] > x)) or ((pointM[0] > x) and (pointN[0] < x)):
                pointS = calculate_intersection(pointK, pointL, pointM, pointN)
                intersections.append(pointS[1])

            i = i + 1

        if filling_orientation:
            intersections.sort(reverse = False)

            line = []
            point = []
            point.append(x)
            point.append(yMin)
            line.append(point)
            j = 0
            while j < len(intersections):
                point = []
                point.append(x)
                point.append(intersections[j])
                line.append(point)
                fillingpattern.append(line)
                line = []
                point = []
                point.append(x)
                point.append(intersections[j+1])
                line.append(point)

                j = j + 2

            point = []
            point.append(x)
            point.append(yMax)
            line.append(point)
            fillingpattern.append(line)

        else:
            intersections.sort(reverse = True)

            line = []
            point = []
            point.append(x)
            point.append(yMax)
            line.append(point)
            j = 0
            while j < len(intersections):
                point = []
                point.append(x)
                point.append(intersections[j])
                line.append(point)
                fillingpattern.append(line)
                line = []
                point = []
                point.append(x)
                point.append(intersections[j+1])
                line.append(point)

                j = j + 2

            point = []
            point.append(x)
            point.append(yMin)
            line.append(point)
            fillingpattern.append(line)
            
        x = x + distance_fillinglines

        if bidirectional:
            filling_orientation = not filling_orientation  #Richtungswechsel bei jeder Zeile -> Wegoptimierung

    return fillingpattern


def explode_to_lines(all_polylines):
    i = 0
    lines = []
    while i < len(all_polylines):
        j = 0
        while j < len(all_polylines[i].coords)-1:
            line = []
            line.append(all_polylines[i].coords[j])
            if j < len(all_polylines[i].coords)-1:
                line.append(all_polylines[i].coords[j+1])
            else:
                line.append(all_polylines[i].coords[0])
            lines.append(line)
            j = j + 1
        i = i + 1
    return lines


def eliminate_vertical_lines(all_lines):
    i = 0
    lines = deepcopy(all_lines)
    while i < len(lines):
        if lines[i][0][0] == lines[i][1][0]:
            del(lines[i])
            i = -1
        i = i + 1
    return lines


def plot_geometry(geometry, color):
    i = 0
    while i < len(geometry):
        j = 0
        x = []
        y = []
        while j < len(geometry[i].coords):
            x.append(geometry[i].coords[j][0])
            y.append(geometry[i].coords[j][1])

            j = j + 1
        plot_lines(x, y, color)
        print(i),
        i = i + 1
    print("")


def plot_positioning(geometry, color):
    i = 0
    x = []
    y = []
    while i < len(geometry)-1:
        x.append(geometry[i].coords[0][0])
        y.append(geometry[i].coords[0][1])

        i = i + 1

    plot_lines(x, y, color)


# Tested only with KiCAD 4.0.2 gbr-export. No general import function.
def gbr_import(fname, delta):
#    global import_as_perimeter
    file_length = file_len(fname)
    line = ""
    tools = []
    geometry = []
    x_pos_old = 0
    y_pos_old = 0
    linie = []
    filled_area = False
    filled_area_delayed = False
    tool_delayed = 0

    print("")
    print("Import tools"),
    print(filename_input)

    file_input = open(fname)

    i = 0
    while ((i < file_length) and (line <> "G04 APERTURE LIST")):
        line = file_input.readline()
        line = line.strip()   
        line = line.split("*")[0]
        i = i + 1

# -----------Import Tools------------
    while ((i < file_length) and (line <> "G04 APERTURE END LIST")):
        line = file_input.readline()
        line = line.strip()
        line = line.split("*")[0]

        if line.find(",", 0, len(line)) == 7:
            tool=[]
#           Toolname:
            tool.append(line[4:6])
#           Toolshape:
            tool.append(line[6:7])
#           Toolsize:
            line = line.split(",")[1]
            if line.find("X", 0, len(line)) == -1:
                tool.append(float(line))
            else:
                tool.append(float(line.split("X")[0]))
                tool.append(float(line.split("X")[1]))

            print(tool)
                
            tools.append(tool)
                    
        i = i + 1

# -----------Import geometry------------
    if len(tools) == 0:
        print("No tools found!")
    else:
        while ((i < file_length) and (line <> "M02*")):
            line = file_input.readline()
            line = line.strip()
            line = line.split("*")[0]

            if line[0:1] == "G":
                filled_area = True

            if line[0:1] == "D":
                toolname = line[1:3]
                filled_area = False
                
                j = 0
                while j < len(tools):
                    if toolname == tools[j][0]:
                        tool = tools[j]
                    j = j + 1

#            if line[0:1] == "X":
#                x_string = line.split("-")[0]
#                x_string = ''.join(i for i in x_string if i.isdigit())
#                x_float = float(x_string) / 1000000
#                
#                y_string = line.split("-")[1]
#                y_string = y_string.split("D")[0]
#                y_string = ''.join(i for i in y_string if i.isdigit())
#                y_float = float(y_string) / 1000000
#
#                action_string = line.split("-")[1]
#                action_string = action_string.split("D")[1]

            if line[0:1] == "X":
                x_string = line.split("Y")[0]
                x_string = x_string.split("X")[1]
                x_float = float(x_string) / 1000000
                
                y_string = line.split("Y")[1]
                y_string = y_string.split("D")[0]
                y_float = float(y_string) / 1000000
#                y_float = float(y_string) / 1000000 * (-1)

                action_string = line.split("D")[1]
                action_string = action_string.split("*")[0]

#                print("X: "),
#                print(x_float),
#                print(" Y: "),
#                print(y_float),
#                print(" Action: "),
#                print(action_string)

                if action_string == "03":
#                    print(tool)
                    if tool[1] == "R":
                        polygon = Polygon([(x_float -(tool[2]/2 + delta), y_float -(tool[3]/2 + delta)), (x_float -(tool[2]/2 + delta), y_float +(tool[3]/2 + delta)), (x_float +(tool[2]/2 + delta), y_float +(tool[3]/2 + delta)), (x_float +(tool[2]/2 + delta), y_float -(tool[3]/2 + delta)), (x_float -(tool[2]/2 + delta), y_float -(tool[3]/2 + delta))])                    
                    if tool[1] == "C":
                        polygon = Point(x_float, y_float).buffer(tool[2]/2 + delta)
                    if tool[1] == "O":
                        polygon = oval(x_float, y_float, tool[2]/2 + delta, tool[3]/2 + delta)
                    geometry.append(polygon)

                if action_string == "02":
                    if linie <> []:
                        if filled_area_delayed == False:
                            polygon = asLineString(linie).buffer(tool_delayed)
                            geometry.append(polygon)
                        else:  
                            polygon = asPolygon(linie)
                            geometry.append(polygon.buffer(tool_delayed))

                    filled_area_delayed = filled_area

                    tool_delayed = tool[2]/2 + delta
                       
                    linie = []
                    punkt = [x_float, y_float]
                    linie.append(punkt)

                if action_string == "01":
                    punkt = [x_float, y_float]
                    linie.append(punkt)                

            i = i + 1

    if linie <> []:
        if filled_area_delayed == False:
            polygon = asLineString(linie).buffer(tool_delayed)
            geometry.append(polygon)
        else:
            polygon = asPolygon(linie)
            geometry.append(polygon.buffer(tool_delayed))

    return MultiPolygon(geometry)


def oval(x_pos, y_pos, x_radius, y_radius):
    polygon = []
    if (x_radius > y_radius):
        polygon = Polygon([(x_pos - (x_radius - y_radius), y_pos - y_radius), (x_pos + (x_radius - y_radius), y_pos - y_radius), (x_pos + (x_radius - y_radius), y_pos + y_radius), (x_pos - (x_radius - y_radius), y_pos + y_radius), (x_pos - (x_radius - y_radius), y_pos - y_radius)])
        polygon = polygon.union(Point(x_pos - (x_radius - y_radius), y_pos).buffer(y_radius))
        polygon = polygon.union(Point(x_pos + (x_radius - y_radius), y_pos).buffer(y_radius))
    else:
        polygon = Polygon([(x_pos - x_radius, y_pos - (y_radius - x_radius)), (x_pos + x_radius, y_pos - (y_radius - x_radius)), (x_pos + x_radius, y_pos + (y_radius - x_radius)), (x_pos - x_radius, y_pos + (y_radius - x_radius)), (x_pos - x_radius, y_pos - (y_radius - x_radius))])
        polygon = polygon.union(Point(x_pos, y_pos - (y_radius - x_radius)).buffer(x_radius))
        polygon = polygon.union(Point(x_pos, y_pos + (y_radius - x_radius)).buffer(x_radius))

    return polygon


def plot_line(line, color):
    global ax
    x = [line[0][0], line[1][0]]
    y = [line[0][1], line[1][1]]
    ax.plot(x, y, color)


def plot_lines(x, y, color):
    global ax
    ax.plot(x, y, color)


def pyplot_setup(xMin, xMax, yMin, yMax):
    global ax
    fig = plt.figure()
    ax = fig.add_subplot(111)

    xrange = [xMin, xMax]
    yrange = [yMin, yMax]
    ax.set_xlim(*xrange)
    ax.set_ylim(*yrange)
    ax.set_aspect(1)


# Returns the length of a file
def file_len(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
        f.close()
    return i + 1


def calculate_intersection(pointK, pointL, pointM, pointN):
    pointS=[]
    pointS.append(((pointN[0]-pointM[0])*(pointL[0]*pointK[1]-pointK[0]*pointL[1])-(pointL[0]-pointK[0])*(pointN[0]*pointM[1]-pointM[0]*pointN[1]))/((pointN[1]-pointM[1])*(pointL[0]-pointK[0])-(pointL[1]-pointK[1])*(pointN[0]-pointM[0])))
    pointS.append(((pointK[1]-pointL[1])*(pointN[0]*pointM[1]-pointM[0]*pointN[1])-(pointM[1]-pointN[1])*(pointL[0]*pointK[1]-pointK[0]*pointL[1]))/((pointN[1]-pointM[1])*(pointL[0]-pointK[0])-(pointL[1]-pointK[1])*(pointN[0]-pointM[0])))
    return pointS



#---------------------------------------------------------------------------

read_config()

ask_config()

delta = 0.0 * laser_diameter
geometry = gbr_import(filename_input, delta)
geometry =  cascaded_union(geometry).boundary

delta = 0.5 * laser_diameter
perimeter = gbr_import(filename_input, delta)
perimeter =  cascaded_union(perimeter).boundary

drills = import_drill(filename_drl_input)

#xMin = int(geometry.bounds[0] - border)
#yMin = int(geometry.bounds[1] - border)
#xMax = int(geometry.bounds[2] + border)
#yMax = int(geometry.bounds[3] + border)

xMin = geometry.bounds[0] - border
yMin = geometry.bounds[1] - border
xMax = geometry.bounds[2] + border
yMax = geometry.bounds[3] + border

print("")
print("xMin: "),
print(xMin),
print("mm")
print("yMin: "),
print(yMin),
print("mm")
print("xMax: "),
print(xMax),
print("mm")
print("yMAx: "),
print(yMax),
print("mm")

if set_xMin_yMin_to_zero:
    drills = translate_drills(drills, xMin * (-1), yMin * (-1))
#    print(drills)
    geometry = translate(geometry, xMin * (-1), yMin * (-1))
    perimeter = translate(perimeter, xMin * (-1), yMin * (-1))
    xMax = xMax - xMin
    xMin = 0
    yMax = yMax - yMin
    yMin = 0

print("")
print("xMin: "),
print(xMin),
print("mm")
print("yMin: "),
print(yMin),
print("mm")
print("xMax: "),
print(xMax),
print("mm")
print("yMAx: "),
print(yMax),
print("mm")

print("Area: "),
print(int((xMax-xMin)*(yMax-yMin))),
print("mm2")

pyplot_setup(xMin, xMax, yMin, yMax)

print("")
print("Plotting geometry")
plot_geometry(geometry, 'black')


if action_number == 1:

    perimeter = optimizing_outline_travel_V2(perimeter)
    #perimeter = optimizing_outline_arrangement(perimeter)

    print("")
    print("Plotting positioning")
    plot_positioning(perimeter, 'green')
    
    print("")
    print("Plotting perimeter")
    plot_geometry(perimeter, 'red')

    filling_pattern = calculate_filling_pattern(eliminate_vertical_lines(explode_to_lines(perimeter)), bidirectional)

    del_short_lines(filling_pattern, laser_diameter)
    shrink_lines(filling_pattern, laser_diameter)

    if optimize_fillingpattern:
        print("")
        print("Optimizing fillingpattern travel")
        optimize_fillingpattern_positioning(filling_pattern)

    print("")
    print("Plotting fillingpattern")
    plot_fillingpattern(filling_pattern, 'red')
    plot_fillingpattern_positioning(filling_pattern, 'blue')

    export_perimeter = True
    export_filling = True


if action_number == 2:

    filling_pattern = calculate_filling_pattern(eliminate_vertical_lines(explode_to_lines(perimeter)), bidirectional)

    if optimize_fillingpattern:
        print("")
        print("Optimizing fillingpattern travel")
        optimize_fillingpattern_positioning(filling_pattern)

    print("")
    print("Plotting fillingpattern")
    plot_fillingpattern(filling_pattern, 'red')
    plot_fillingpattern_positioning(filling_pattern, 'blue')

    export_perimeter = False
    export_filling = True


if action_number == 3:

    perimeter = optimizing_outline_travel_V2(perimeter)
    #perimeter = optimizing_outline_arrangement(perimeter)

    print("")
    print("Plotting positioning")
    plot_positioning(perimeter, 'green')
    
    print("")
    print("Plotting perimeter")
    plot_geometry(perimeter, 'red')

    filling_pattern = calculate_filling_pattern(eliminate_vertical_lines(explode_to_lines(perimeter)), bidirectional)

    export_perimeter = True
    export_filling = False


export_gcode(perimeter, filling_pattern, export_perimeter, export_filling)
export_x_inv_gcode(perimeter, filling_pattern, export_perimeter, export_filling)
export_y_inv_gcode(perimeter, filling_pattern, export_perimeter, export_filling)

export_drills(drills)
export_x_inv_drills(drills)
export_y_inv_drills(drills)

plot_drills(drills, 'orange')
    
write_config()

plt.show()
